#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;
use iANT::Cfg;
use iANT::Common;
require "iANT/Utils.pl";
use MIME::Base64;
use File::Temp;
use Cwd 'abs_path';

my $O_PARAM;

MAIN:
{

    my $create_index = 0;

    $O_PARAM = New ParamParser(
                               'GETOPTLONG',
                               (
                                "parser_cfg=s", "store_dir=s",
                                "input=s",      "escape_special_characters",
                                "maxdepth=i",   "mindepth=i",
                                "ctime=i",      "verbose",
                                "no_update"
                                 )
                                 );
    $O_PARAM->SetUsage(my $usage = sub {&Usage();});

    $O_PARAM->AssertFileExists('parser_cfg');

    my $parser_cfg = $O_PARAM->Get('parser_cfg');

    $O_PARAM->AssertDefined('store_dir');
    $O_PARAM->AssertDefined('input');

    my $o_parser     = New ParamParser($parser_cfg);
    my $analyzer_mod = $o_parser->Get('analyzer');
    my $analyzer     = new $analyzer_mod;

    $O_PARAM->SetBehaviour('use_substitution_table');
    $O_PARAM = &iANT::Cfg::Load($O_PARAM);

    my $unix_find = $O_PARAM->Get('unix_find');
    my $unix_cat  = $O_PARAM->Get('unix_cat');

    #est-ce que ce sera forcement un project ? non
    #parametre lucene_dir

    my $store_dir = $O_PARAM->Get('store_dir');

    if (-d $store_dir)
    {
        opendir(ACCESSDIR, $store_dir);
        my @a_store_files = readdir(ACCESSDIR);
        closedir(ACCESSDIR);

        #Even if the directory is empty, . and .. exists, so the array has a minimal length of 2
        $create_index = 1 if (@a_store_files == 2);
        if (!$create_index)
        {
            my $nb_cfs = grep (/\.cfs/, @a_store_files);
            if ($nb_cfs == 0)
            {
                print STDERR "ERROR: There are other files than .cfs in the directory\n"
                  if ($O_PARAM->IsDefined('verbose'));
                die("No cfs file in this directory");
            }
        }
    }

    $create_index = 1 if (!-d $store_dir);

    print STDERR "create_index $create_index $store_dir\n" if ($O_PARAM->IsDefined('verbose'));
    mkdir $store_dir if ($create_index);

    my $store = Lucene::Store::FSDirectory->getDirectory($store_dir, $create_index);

    #$doc serait une entree ?
    my @a_files = ();
    if (-d $O_PARAM->Get('input'))
    {
        my $data_dir  = $O_PARAM->Get('input');
        my $extension = $o_parser->Get('suffix');
        print STDERR "EXTENSION: $extension\n\n" if ($O_PARAM->IsDefined('verbose'));

        my $maxdepth = '';
        my $mindepth = '';
        my $ctime    = '';

        $maxdepth = '-maxdepth ' . $O_PARAM->Get('maxdepth') if ($O_PARAM->IsDefined('maxdepth'));
        $mindepth = '-mindepth ' . $O_PARAM->Get('mindepth') if ($O_PARAM->IsDefined('mindepth'));
        $ctime    = '-ctime ' . $O_PARAM->Get('ctime')       if ($O_PARAM->IsDefined('ctime'));

        my $cmd = "$unix_find $data_dir $maxdepth $mindepth -type f -name \\*.$extension";
        print STDERR "CMD: $cmd\n\n" if ($O_PARAM->IsDefined('verbose'));
        @a_files = `$cmd`;
    }
    elsif (-e $O_PARAM->Get('input'))
    {
        if ($O_PARAM->IsDefined('verbose'))
        {
            print STDERR 'WARNING: "--maxdepth" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('maxdepth'));
            print STDERR 'WARNING: "--mindepth" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('mindepth'));
            print STDERR 'WARNING: "--ctime" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('ctime'));
        }

        my $ra_files = &iANT::Common::GetFileNamesFromFof($O_PARAM->Get('input'));

        if ($ra_files)
        {
            @a_files = @{$ra_files};
            print STDERR "INFO: Dealing with a file of files\n" if ($O_PARAM->IsDefined('verbose'));
            1;
        }
        else
        {
            print STDERR "INFO: Dealing with an txt file\n" if ($O_PARAM->IsDefined('verbose'));

            @a_files = ($O_PARAM->Get('input'));
        }

    }

    chomp @a_files;

    my ($ra_fields_to_parse, $rh_docid) = &GetLuceneFields4Txt($o_parser);

    if ((!$create_index) && (!$O_PARAM->IsDefined('no_update')))
    {
        
        #SI DOCID est un element de ligne , alors faut faire du ligne a ligne
        #sinon juste le contenu du fichier

        my $search_query     = '';
        my @a_search_queries = ();

        #utiliser un @queries contenant au max 32 clauses. Limitation Lucene 1.4 a priori ...
        #%32
        my $count            = 0;
        my %h_search_queries = ();
        foreach my $to_add (@a_files)
        {

            if (-B $to_add)
            {
                print STDERR "ERROR: $to_add is a binary file !\n";
                next;
            }        

            my $abs_path = abs_path($to_add);
            my $file;
            my @a_lines = ();

            print STDERR "INFO:read $to_add\n" if ($O_PARAM->IsDefined('verbose'));
            my $fh_to_add = new IO::File($to_add);
            if (defined $fh_to_add)
            {
                while (my $line = <$fh_to_add>)
                {
                    $file .= $line;
                    chomp $line;
                    push(@a_lines, $line);
                }
                $fh_to_add->close;

            }
            else
            {
                print STDERR "ERROR: $to_add cannot be read !\n";
                next;
            }

            if ($rh_docid->{regexp} =~ /^\$line=/)
            {
                foreach my $line (@a_lines)
                {
                    &DeleteDocument($store,$rh_docid, $abs_path, $file, $line);
                }
            }
            else
            {
                &DeleteDocument($store,$rh_docid, $abs_path, $file, undef);
            }


        }

        my $reader;

    }

    my $writer = new Lucene::Index::IndexWriter($store, $analyzer, $create_index);
    $writer->optimize();

    foreach my $to_add (@a_files)
    {

        if (-B $to_add)
        {
            print STDERR "ERROR: $to_add is a binary file !\n";
            next;
        }

        my $abs_path = abs_path($to_add);
        my $file;
        my @a_lines = ();

        print STDERR "INFO:read $to_add\n" if ($O_PARAM->IsDefined('verbose'));
        my $fh_to_add = new IO::File($to_add);
        if (defined $fh_to_add)
        {
            while (my $line = <$fh_to_add>)
            {
                $file .= $line;
                chomp $line;
                push(@a_lines, $line);
            }
            $fh_to_add->close;

        }
        else
        {
            print STDERR "ERROR: $to_add cannot be read !\n";
            next;
        }

        print STDERR "INFO:add $to_add\n" if ($O_PARAM->IsDefined('verbose'));
        if (&LineByLineAnalysis($ra_fields_to_parse) == 1)
        {
            print STDERR "INFO:line by line\n" if ($O_PARAM->IsDefined('verbose'));
            foreach my $line (@a_lines)
            {
                &AddDocument($writer, $ra_fields_to_parse,$rh_docid, $abs_path, $file, $line);
            }
        }
        else
        {
            &AddDocument($writer, $ra_fields_to_parse,$rh_docid, $abs_path, $file, undef);
        }

        # field gets just stored
        #tout le document txt ?

    }

    $writer->optimize();
    $writer->close();
    undef $writer;

}

sub LineByLineAnalysis
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        my $regexp = $rh_field_to_parse->{regexp};
        return 1 if ($regexp =~ /^\$line=/);
    }
    return 0;
}

sub DeleteDocument
{

    my ($store,$rh_docid, $abs_path, $file, $line) = @_;
    
    my ($docid_value) = eval($rh_docid->{regexp});
    $docid_value = &EscapeLuceneSpecialCharaters($docid_value) if ($O_PARAM->IsDefined('escape_special_characters'));
    
    
    if ((defined $docid_value) && ($docid_value ne ''))
    {
        my $reader;
        eval {$reader = Lucene::Index::IndexReader->open($store)};

        if ($@)
        {
            print STDERR $@;
        }
        else
        {

            $reader = Lucene::Index::IndexReader->open($store);

            #if indexer = Text ---> lc($value) mode
            #le terme Lucene est mis en minuscule quand l'indexer est du texte
            #il n'y a que l'indexer Keyword qui ne change pas la casse

            $docid_value = lc($docid_value) if ($rh_docid->{mode} !~ /keyword/i);

            print STDOUT "INFO: Delete entries with term " . $rh_docid->{alias} . " '$docid_value'\n" if ($O_PARAM->IsDefined('verbose'));
            my $term = new Lucene::Index::Term($rh_docid->{alias}, $docid_value);
            $reader->deleteDocuments($term) or print "ERROR Delete " . $rh_docid->{alias} . ' ' . $docid_value . "\n";
            $reader->close();
            undef $reader;

        }
    }
    return;
}


sub AddDocument
{
    my ($writer, $ra_fields_to_parse,$rh_docid, $abs_path, $file, $line) = @_;
    my $unix_gzip = $O_PARAM->Get('unix_gzip');

    my $doc = new Lucene::Document;
    my $doc_id_value;
    
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        my $regexp = $rh_field_to_parse->{regexp};
        my $alias  = $rh_field_to_parse->{alias};
        my $mode   = $rh_field_to_parse->{mode};
        my $value;
        
        if ($regexp eq 'file')
        {

            #gzip + b64 => compression d'un facteur 2 a 2.5

            my $tmp_dir = $O_PARAM->Get('tmp_dir');
            my ($tmp_fh, $tmp_filename) = File::Temp::tempfile(DIR => $tmp_dir);
            my $gz_txt = `$unix_gzip -c $abs_path`;
            chomp $gz_txt;
            unlink($tmp_filename);
            $value = MIME::Base64::encode($gz_txt);
            $doc->add(Lucene::Document::Field->$mode("$alias", "$value"));
            undef $value;
        }
        else
        {
            my @a_value = eval($regexp);

            if ($@)
            {
                print STDERR "Error in $regexp: $@\n";
            }
            else
            {
                if ($#a_value > 0)
                {
                    $value = join(' // ', @a_value);
                }
                else
                {
                    $value = $a_value[0];
                }

            }
        }
        
        $doc_id_value = $value if ($alias eq $rh_docid->{alias});
        
        if (defined $value && ($value ne ''))
        {
            if ($mode =~ /unindexed/i)
            {
                my $tmp_dir = $O_PARAM->Get('tmp_dir');
                my ($tmp_fh, $tmp_filename) = File::Temp::tempfile(DIR => $tmp_dir);
                print $tmp_fh $value;
                my $gz_txt = `$unix_gzip -c $tmp_filename`;
                chomp $gz_txt;
                unlink($tmp_filename);
                $value = MIME::Base64::encode($gz_txt);
            }
            $doc->add(Lucene::Document::Field->$mode("$alias", $value));
        }
    }
    
    if (defined $doc_id_value && ($doc_id_value ne ''))
    {
        $writer->addDocument($doc);
    }
    else
    {
        print STDERR "WARNING a document in $abs_path has not been added because is docid is empty\n";
    }   

}

sub Usage
{

    print STDOUT <<END;
   
Usage:

   $0 --parser_cfg=<file> --store_dir=<directory> --input=<file|directory> [--maxdepth=i] [--mindepth=i] [--escape_special_characters]

Description

    Add/Update text files to a Lucene index
   
 Mandatory
    --parser_cfg:   parser configuration file (contains Analyzer name, fields to parse, ...)
    --store_dir :   directory where Lucene indeces are stored. If you want to create a new index, the store repository must not exist, lucene is going to create it.
    --input     :   file : add file to index
                    directory: scan directory looking for files with extension defined in 
                    parser_cfg file (parameter:suffix)
		    
  Optional
     --escape_special_characters : to escape special characters in Lucene query
			           list: \ + - ! ( ) { } [ ] ^ " ~ * ? :

     --maxdepth	:   [unix find option] Descend at most "i" levels of directories below the command line arguments (--input is a directory)
     --mindepth	:   [unix find option] Do  not apply any tests or actions at levels less than "i" (--input is a directory)
     --ctime	:   [unix find option] Files status was last changed i*24 hours ago (--input is a directory)
     --no_update:   Doesn't check if a entry belongs to index. Add it anyway.
                                

END
}
